@@ -28,50 +28,21 @@
 #endif
 #define TRACK_MAX_FRAMES ULONG_MAX/4/1152
 
+#include "mpeghead.h"
+
 #include "debug.h"
 
 #define bsbufid(fr) (fr)->bsbuf==(fr)->bsspace[0] ? 0 : ((fr)->bsbuf==fr->bsspace[1] ? 1 : ( (fr)->bsbuf==(fr)->bsspace[0]+512 ? 2 : ((fr)->bsbuf==fr->bsspace[1]+512 ? 3 : -1) ) )
 
-/*
-	AAAAAAAA AAABBCCD EEEEFFGH IIJJKLMM
-	A: sync
-	B: mpeg version
-	C: layer
-	D: CRC
-	E: bitrate
-	F: sampling rate
-	G: padding
-	H: private
-	I: channel mode
-	J: mode ext
-	K: copyright
-	L: original
-	M: emphasis
-
-	old compare mask 0xfffffd00:
-	11111111 11111111 11111101 00000000
-
-	means: everything must match excluding padding and channel mode, ext mode, ...
-	But a vbr stream's headers will differ in bitrate!
-	We are already strict in allowing only frames of same type in stream, we should at least watch out for VBR while being strict.
-
-	So a better mask is:
-	11111111 11111111 00001101 00000000
-
-	Even more, I'll allow varying crc bit.
-	11111111 11111110 00001101 00000000
-
-	Still unsure about this private bit... well, as Marcel pointed out:
-	The decoder should not care about this.
-	11111111 11111110 00001100 00000000
-*/
-#define HDRCMPMASK 0xfffe0c00
-/*
-	This needs cleanup ... The parser is not as strict as documented above.
-	Here comes a mask that checks for anything that can change sampling rate.
-	This includes change of MPEG version and frequency bits.
-	00011000 00001100 00000000, BB and FF bits */
-#define HDRSAMPMASK 0x180c00
+enum parse_codes
+{
+	 PARSE_MORE = MPG123_NEED_MORE
+	,PARSE_ERR  = MPG123_ERR
+	,PARSE_END  = 0 /* No more audio data to find. */
+	,PARSE_GOOD = 1 /* Everything's fine. */
+	,PARSE_RESYNC = 2 /* Header not good, go into resync. */
+	,PARSE_AGAIN  = 3 /* Really start over, throw away and read a new header, again. */
+};
 
 /* bitrates for [mpeg1/2][layer] */
 static const int tabsel_123[2][3][16] =
@@ -91,6 +62,9 @@ static const int tabsel_123[2][3][16] =
 static const long freqs[9] = { 44100, 48000, 32000, 22050, 24000, 16000 , 11025 , 12000 , 8000 };
 
 static int decode_header(mpg123_handle *fr,unsigned long newhead);
+static int skip_junk(mpg123_handle *fr, unsigned long *newheadp, long *headcount);
+static int do_readahead(mpg123_handle *fr, unsigned long newhead);
+static int wetwork(mpg123_handle *fr, unsigned long *newheadp);
 
 /* These two are to be replaced by one function that gives all the frame parameters (for outsiders).*/
 /* Those functions are unsafe regarding bad arguments (inside the mpg123_handle), but just returning anything would also be unsafe, the caller code has to be trusted. */
@@ -105,24 +79,21 @@ long frame_freq(mpg123_handle *fr)
 	return freqs[fr->sampling_frequency];
 }
 
-#define free_format_header(head) ( ((head & 0xffe00000) == 0xffe00000) && ((head>>17)&3) && (((head>>12)&0xf) == 0x0) && (((head>>10)&0x3) != 0x3 ))
-
 /* compiler is smart enought to inline this one or should I really do it as macro...? */
 static int head_check(unsigned long head)
 {
 	if
 	(
-		/* first 11 bits are set to 1 for frame sync */
-		((head & 0xffe00000) != 0xffe00000)
+		((head & HDR_SYNC) != HDR_SYNC)
 		||
 		/* layer: 01,10,11 is 1,2,3; 00 is reserved */
-		(!((head>>17)&3))
+		(!(HDR_LAYER_VAL(head)))
 		||
 		/* 1111 means bad bitrate */
-		(((head>>12)&0xf) == 0xf)
+		(HDR_BITRATE_VAL(head) == 0xf)
 		||
 		/* sampling freq: 11 is reserved */
-		(((head>>10)&0x3) == 0x3 )
+		(HDR_SAMPLERATE_VAL(head) == 0x3)
 		/* here used to be a mpeg 2.5 check... re-enabled 2.5 decoding due to lack of evidence that it is really not good */
 	)
 	{
@@ -147,14 +118,10 @@ static int check_lame_tag(mpg123_handle *fr)
 		I hope that ensuring all zeros until tag start is enough.
 	*/
 	int lame_offset = (fr->stereo == 2) ? (fr->lsf ? 17 : 32 ) : (fr->lsf ? 9 : 17);
-	/* At least skip the decoder delay. */
-#ifdef GAPLESS
-	if(fr->p.flags & MPG123_GAPLESS)
-	{
-		if(fr->begin_s == 0) frame_gapless_init(fr, GAPLESS_DELAY, 0);
-	}
-#endif
 
+	if(fr->p.flags & MPG123_IGNORE_INFOFRAME) return 0;
+
+	/* Note: CRC or not, that does not matter here. */
 	if(fr->framesize >= 120+lame_offset) /* traditional Xing header is 120 bytes */
 	{
 		int i;
@@ -219,15 +186,11 @@ static int check_lame_tag(mpg123_handle *fr)
 						*/
 						fr->track_frames = (off_t) make_long(fr->bsbuf, lame_offset);
 						if(fr->track_frames > TRACK_MAX_FRAMES) fr->track_frames = 0; /* endless stream? */
-						#ifdef GAPLESS
-						/* if no further info there, remove/add at least the decoder delay */
+#ifdef GAPLESS
+						/* if no further info there, remove at least the decoder delay */
 						if(fr->p.flags & MPG123_GAPLESS)
-						{
-							off_t length = fr->track_frames * spf(fr);
-							if(length > 1)
-							frame_gapless_init(fr, GAPLESS_DELAY, length+GAPLESS_DELAY);
-						}
-						#endif
+						frame_gapless_init(fr, fr->track_frames, 0, 0);
+#endif
 						if(VERBOSE3) fprintf(stderr, "Note: Xing: %lu frames\n", (long unsigned)fr->track_frames);
 					}
 
@@ -282,21 +245,35 @@ static int check_lame_tag(mpg123_handle *fr)
 					lame_offset += 4;
 				}
 				/* I guess that either 0 or LAME extra data follows */
-				/* there may this crc16 be floating around... (?) */
 				if(fr->bsbuf[lame_offset] != 0)
 				{
 					unsigned char lame_vbr;
 					float replay_gain[2] = {0,0};
 					float peak = 0;
-					/* float gain_offset = 0; */ /* going to be +6 for old lame that used 83dB */
+					float gain_offset = 0; /* going to be +6 for old lame that used 83dB */
 					char nb[10];
 					memcpy(nb, fr->bsbuf+lame_offset, 9);
 					nb[9] = 0;
 					if(VERBOSE3) fprintf(stderr, "Note: Info: Encoder: %s\n", nb);
 					if(!strncmp("LAME", nb, 4))
 					{
-						/* gain_offset = 6; */
-						debug("TODO: finish lame detetcion...");
+						/* Lame versions before 3.95.1 used 83 dB reference level, later versions 89 dB.
+						   We stick with 89 dB as being "normal", adding 6 dB. */
+						unsigned int major, minor;
+						char rest[6];
+						rest[0] = 0;
+						if(sscanf(nb+4, "%u.%u%s", &major, &minor, rest) >= 2)
+						{
+							debug3("LAME: %u/%u/%s", major, minor, rest);
+							/* We cannot detect LAME 3.95 reliably (same version string as 3.95.1), so this is a blind spot.
+							   Everything < 3.95 is safe, though. */
+							if(major < 3 || (major == 3 && minor < 95))  /* || (major == 3 && minor == 95 && rest[0] == 0)) */
+							{
+								gain_offset = 6;
+								if(VERBOSE3) fprintf(stderr, "Note: Info: Old LAME detected, using ReplayGain preamp of %f dB.\n", gain_offset);
+							}
+						}
+						else if(VERBOSE3) fprintf(stderr, "Note: Info: Cannot determine LAME version.\n");
 					}
 					lame_offset += 9;
 					/* the 4 big bits are tag revision, the small bits vbr method */
@@ -357,6 +334,9 @@ static int check_lame_tag(mpg123_handle *fr)
 							else continue;
 							/* get the 9 bits into a number, divide by 10, multiply sign... happy bit banging */
 							replay_gain[gt] = (float) ((fr->bsbuf[lame_offset] & 0x2) ? -0.1 : 0.1) * (make_short(fr->bsbuf, lame_offset) & 0x1ff);
+							/* If this is an automatic value from LAME (or whatever), the automatic gain offset applies.
+							   If a user or whoever set the value, do not touch it! 011 is automatic origin. */
+							if(origin == 3) replay_gain[gt] += gain_offset;
 						}
 						lame_offset += 2;
 					}
@@ -389,13 +369,9 @@ static int check_lame_tag(mpg123_handle *fr)
 					#ifdef GAPLESS
 					if(fr->p.flags & MPG123_GAPLESS)
 					{
-						off_t length = fr->track_frames * spf(fr);
-						off_t skipbegin = GAPLESS_DELAY + ((((int) fr->bsbuf[lame_offset]) << 4) | (((int) fr->bsbuf[lame_offset+1]) >> 4));
-						off_t skipend = -GAPLESS_DELAY + (((((int) fr->bsbuf[lame_offset+1]) << 8) | (((int) fr->bsbuf[lame_offset+2]))) & 0xfff);
-						debug3("preparing gapless mode for layer3: length %lu, skipbegin %lu, skipend %lu", 
-								(long unsigned)length, (long unsigned)skipbegin, (long unsigned)skipend);
-						if(length > 1)
-						frame_gapless_init(fr, skipbegin, (skipend < length) ? length-skipend : length);
+						off_t skipbegin = ((((int) fr->bsbuf[lame_offset]) << 4) | (((int) fr->bsbuf[lame_offset+1]) >> 4));
+						off_t skipend = (((((int) fr->bsbuf[lame_offset+1]) << 8) | (((int) fr->bsbuf[lame_offset+2]))) & 0xfff);
+						frame_gapless_init(fr, fr->track_frames, skipbegin, skipend);
 					}
 					#endif
 				}
@@ -412,31 +388,22 @@ static int check_lame_tag(mpg123_handle *fr)
 /* Just tell if the header is some mono. */
 static int header_mono(unsigned long newhead)
 {
-	return ((newhead>>6)&0x3) == MPG_MD_MONO ? TRUE : FALSE;
+	return HDR_CHANNEL_VAL(newhead) == MPG_MD_MONO ? TRUE : FALSE;
 }
 
-/*
-	That's a big one: read the next frame. 1 is success, <= 0 is some error
-	Special error READER_MORE means: Please feed more data and try again.
-*/
-int read_frame(mpg123_handle *fr)
+static void halfspeed_prepare(mpg123_handle *fr)
 {
-	/* TODO: rework this thing */
-	unsigned long newhead;
-	off_t framepos;
-	int ret;
-	off_t oret;
-	/* stuff that needs resetting if complete frame reading fails */
-	int oldsize  = fr->framesize;
-	int oldphase = fr->halfphase;
-
-	/* The counter for the search-first-header loop.
-	   It is persistent outside the loop to prevent seemingly endless loops
-	   when repeatedly headers are found that do not have valid followup headers. */
-	int headcount = 0;
-
-	fr->fsizeold=fr->framesize;       /* for Layer3 */
+	/* save for repetition */
+	if(fr->p.halfspeed && fr->lay == 3)
+	{
+		debug("halfspeed - reusing old bsbuf ");
+		memcpy (fr->ssave, fr->bsbuf, fr->ssize);
+	}
+}
 
+/* If this returns 1, the next frame is the repetition. */
+static int halfspeed_do(mpg123_handle *fr)
+{
 	/* Speed-down hack: Play it again, Sam (the frame, I mean). */
 	if (fr->p.halfspeed) 
 	{
@@ -456,6 +423,43 @@ int read_frame(mpg123_handle *fr)
 			fr->halfphase = fr->p.halfspeed - 1;
 		}
 	}
+	return 0;
+}
+
+/* 
+	Temporary macro until we got this worked out.
+	Idea is to filter out special return values that shall trigger direct jumps to end / resync / read again. 
+	Particularily, ret==PARSE_END==0 and ret==PARSE_GOOD==1 are not affected.
+*/
+#define JUMP_CONCLUSION(ret) \
+{ \
+if(ret < 0){ debug1("%s", ret == MPG123_NEED_MORE ? "need more" : "read error"); goto read_frame_bad; } \
+else if(ret == PARSE_AGAIN) goto read_again; \
+else if(ret == PARSE_RESYNC) goto init_resync; \
+}
+
+/*
+	That's a big one: read the next frame. 1 is success, <= 0 is some error
+	Special error READER_MORE means: Please feed more data and try again.
+*/
+int read_frame(mpg123_handle *fr)
+{
+	/* TODO: rework this thing */
+	unsigned long newhead;
+	off_t framepos;
+	int ret;
+	/* stuff that needs resetting if complete frame reading fails */
+	int oldsize  = fr->framesize;
+	int oldphase = fr->halfphase;
+
+	/* The counter for the search-first-header loop.
+	   It is persistent outside the loop to prevent seemingly endless loops
+	   when repeatedly headers are found that do not have valid followup headers. */
+	long headcount = 0;
+
+	fr->fsizeold=fr->framesize;       /* for Layer3 */
+
+	if(halfspeed_do(fr) == 1) return 1;
 
 read_again:
 	/* In case we are looping to find a valid frame, discard any buffered data before the current position.
@@ -463,7 +467,7 @@ read_again:
 	if(fr->rd->forget != NULL) fr->rd->forget(fr);
 
 	debug2("trying to get frame %"OFF_P" at %"OFF_P, (off_p)fr->num+1, (off_p)fr->rd->tell(fr));
-	if((ret = fr->rd->head_read(fr,&newhead)) <= 0){ debug("need more?"); goto read_frame_bad;}
+	if((ret = fr->rd->head_read(fr,&newhead)) <= 0){ debug1("need more? (%i)", ret); goto read_frame_bad;}
 
 init_resync:
 
@@ -474,7 +478,7 @@ init_resync:
 		else
 		/* If they have the same sample rate. Note that only is _not_ the case for the first header, as we enforce sample rate match for following frames.
 			 So, during one stream, only change of stereoness is possible and indicated by header_change == 2. */
-		if((fr->oldhead & HDRSAMPMASK) == (newhead & HDRSAMPMASK))
+		if((fr->oldhead & HDR_SAMPMASK) == (newhead & HDR_SAMPMASK))
 		{
 			/* Now if both channel modes are mono or both stereo, it's no big deal. */
 			if( header_mono(fr->oldhead) == header_mono(newhead))
@@ -483,244 +487,33 @@ init_resync:
 	}
 
 #ifdef SKIP_JUNK
-	/* watch out for junk/tags on beginning of stream by invalid header */
-	if(!fr->firsthead && !head_check(newhead)) {
-
-		/* check for id3v2; first three bytes (of 4) are "ID3" */
-		if((newhead & (unsigned long) 0xffffff00) == (unsigned long) 0x49443300)
-		{
-			int id3ret = 0;
-			id3ret = parse_new_id3(fr, newhead);
-			if     (id3ret < 0){ debug("need more?"); ret = id3ret; goto read_frame_bad; }
-#ifndef NO_ID3V2
-			else if(id3ret > 0){ debug("got ID3v2"); fr->metaflags  |= MPG123_NEW_ID3|MPG123_ID3; }
-			else debug("no useful ID3v2");
-#endif
-
-			fr->oldhead = 0;
-			goto read_again; /* Also in case of invalid ID3 tag (ret==0), try to get on track again. */
-		}
-		else if(VERBOSE2 && fr->silent_resync == 0) fprintf(stderr,"Note: Junk at the beginning (0x%08lx)\n",newhead);
-
-		/* I even saw RIFF headers at the beginning of MPEG streams ;( */
-		if(newhead == ('R'<<24)+('I'<<16)+('F'<<8)+'F') {
-			if(VERBOSE2 && fr->silent_resync == 0) fprintf(stderr, "Note: Looks like a RIFF header.\n");
-
-			if((ret=fr->rd->head_read(fr,&newhead))<=0){ debug("need more?"); goto read_frame_bad; }
-
-			while(newhead != ('d'<<24)+('a'<<16)+('t'<<8)+'a')
-			{
-				if((ret=fr->rd->head_shift(fr,&newhead))<=0){ debug("need more?"); goto read_frame_bad; }
-			}
-			if((ret=fr->rd->head_read(fr,&newhead))<=0){ debug("need more?"); goto read_frame_bad; }
-
-			if(VERBOSE2 && fr->silent_resync == 0) fprintf(stderr,"Note: Skipped RIFF header!\n");
-
-			fr->oldhead = 0;
-			goto read_again;
-		}
-		/* unhandled junk... just continue search for a header */
-		/* step in byte steps through next 64K */
-		debug("searching for header...");
-
-		ret = 0; /* We will check the value after the loop. */
-		for(; headcount<65536; headcount++)
-		{
-			if((ret=fr->rd->head_shift(fr,&newhead))<=0){ debug("need more?"); goto read_frame_bad; }
-			/* if(head_check(newhead)) */
-			if(head_check(newhead) && (ret=decode_header(fr, newhead)))
-				break;
-		}
-		if(ret<0){ debug("need more?"); goto read_frame_bad; }
-
-		if(headcount == 65536)
-		{
-			if(NOQUIET) error("Giving up searching valid MPEG header after (over) 64K of junk.");
-			return 0;
-		}
-		else debug1("hopefully found one at %"OFF_P, (off_p)fr->rd->tell(fr));
-		/* 
-		 * should we additionaly check, whether a new frame starts at
-		 * the next expected position? (some kind of read ahead)
-		 * We could implement this easily, at least for files.
-		 */
-	}
-#endif
-
-	/* first attempt of read ahead check to find the real first header; cannot believe what junk is out there! */
-	if(!fr->firsthead && fr->rdat.flags & (READER_SEEKABLE|READER_BUFFERED) && head_check(newhead) && (ret=decode_header(fr, newhead)))
+	if(!fr->firsthead && !head_check(newhead))
 	{
-		unsigned long nexthead = 0;
-		int hd = 0;
-		off_t start = fr->rd->tell(fr);
-		if(ret<0){ debug("need more?"); goto read_frame_bad; }
-
-		debug2("doing ahead check with BPF %d at %"OFF_P, fr->framesize+4, (off_p)start);
-		/* step framesize bytes forward and read next possible header*/
-		if((oret=fr->rd->skip_bytes(fr, fr->framesize))<0)
-		{
-			if(oret==READER_ERROR && NOQUIET) error("cannot seek!");
-
-			ret = oret; /* Purposeful negative values aren't big. */
-			goto read_frame_bad;
-		}
-		hd = fr->rd->head_read(fr,&nexthead);
-		if(hd==MPG123_NEED_MORE){ debug("need more?"); ret = hd; goto read_frame_bad; }
-		if((ret=fr->rd->back_bytes(fr, fr->rd->tell(fr)-start))<0)
-		{
-			if(ret==READER_ERROR && NOQUIET) error("cannot seek!");
-			else debug("need more?"); 
-			goto read_frame_bad;
-		}
-		debug1("After fetching next header, at %"OFF_P, (off_p)fr->rd->tell(fr));
-		if(!hd)
-		{
-			if(NOQUIET) warning("cannot read next header, a one-frame stream? Duh...");
-		}
-		else
-		{
-			debug2("does next header 0x%08lx match first 0x%08lx?", nexthead, newhead);
-			/* not allowing free format yet */
-			if(!head_check(nexthead) || (nexthead & HDRCMPMASK) != (newhead & HDRCMPMASK))
-			{
-				debug("No, the header was not valid, start from beginning...");
-				fr->oldhead = 0; /* start over */
-				/* try next byte for valid header */
-				if((ret=fr->rd->back_bytes(fr, 3))<0)
-				{
-					if(NOQUIET) error("cannot seek!");
-					else debug("need more?"); 
-					goto read_frame_bad;
-				}
-				goto read_again;
-			}
-		}
+		ret = skip_junk(fr, &newhead, &headcount);
+		JUMP_CONCLUSION(ret);
 	}
+#endif
 
-	/*
-		Why has this head check been avoided here before? And apart from that:
-		Tricky business: Bug 3267863 triggers endless loop because an invalid free format header does not trigger head_check(), hence does not trigger resync.
-		This whole logic is going to be reworked, code untangled, but 1.13.3 needs some minimal patching.
-		So we try to fully decode the header (which includes free format size checking), not just check.
-	*/
-	/* Check in advance to avoid premature error message in decode_header() */
 	ret = head_check(newhead);
-	if(ret)
-	{
-		ret = decode_header(fr, newhead);
-		if(ret < 0) goto read_frame_bad;
+	if(ret) ret = decode_header(fr, newhead);
+
+	JUMP_CONCLUSION(ret); /* That only continues for ret == 0 or 1 */
+	if(ret == 0)
+	{ /* Header was not good. */
+		ret = wetwork(fr, &newhead); /* Messy stuff, handle junk, resync ... */
+		JUMP_CONCLUSION(ret);
+		if(ret == 0) goto read_frame_bad;
 	}
 
-	if(!ret)
+	if(!fr->firsthead)
 	{
-		/* and those ugly ID3 tags */
-		if((newhead & 0xffffff00) == ('T'<<24)+('A'<<16)+('G'<<8))
-		{
-			fr->id3buf[0] = (unsigned char) ((newhead >> 24) & 0xff);
-			fr->id3buf[1] = (unsigned char) ((newhead >> 16) & 0xff);
-			fr->id3buf[2] = (unsigned char) ((newhead >> 8)  & 0xff);
-			fr->id3buf[3] = (unsigned char) ( newhead        & 0xff);
-			if((ret=fr->rd->fullread(fr,fr->id3buf+4,124)) < 0){ debug("need more?"); goto read_frame_bad; }
-			fr->metaflags  |= MPG123_NEW_ID3|MPG123_ID3;
-			fr->rdat.flags |= READER_ID3TAG; /* that marks id3v1 */
-			if (VERBOSE3) fprintf(stderr,"Note: Skipped ID3v1 tag.\n");
-			goto read_again;
-		}
-		/* duplicated code from above! */
-		/* check for id3v2; first three bytes (of 4) are "ID3" */
-		if((newhead & (unsigned long) 0xffffff00) == (unsigned long) 0x49443300)
-		{
-			int id3length = 0;
-			id3length = parse_new_id3(fr, newhead);
-			if(id3length < 0){ debug("need more?"); ret = id3length; goto read_frame_bad; }
-
-			fr->metaflags  |= MPG123_NEW_ID3|MPG123_ID3;
-			goto read_again;
-		}
-		else if(NOQUIET && fr->silent_resync == 0)
-		{
-			fprintf(stderr,"Note: Illegal Audio-MPEG-Header 0x%08lx at offset %"OFF_P".\n",
-				newhead, (off_p)fr->rd->tell(fr)-4);
-		}
-
-		/* We reach this point only for invalid headers.
-		   And: All paths lead to a jump out of this block! So there is no "after". */
-
-		if(NOQUIET && (newhead & 0xffffff00) == ('b'<<24)+('m'<<16)+('p'<<8)) fprintf(stderr,"Note: Could be a BMP album art.\n");
-		/* Do resync if not forbidden by flag.
-		I used to have a check for not-icy-meta here, but concluded that the desync issues came from a reader bug, not the stream. */
-		if( !(fr->p.flags & MPG123_NO_RESYNC) )
-		{
-			long try = 0;
-			long limit = fr->p.resync_limit;
-			
-			/* If a resync is needed the bitreservoir of previous frames is no longer valid */
-			fr->bitreservoir = 0;
-
-			/* TODO: make this more robust, I'd like to cat two mp3 fragments together (in a dirty way) and still have mpg123 beign able to decode all it somehow. */
-			if(NOQUIET && fr->silent_resync == 0) fprintf(stderr, "Note: Trying to resync...\n");
-			/* Read more bytes until we find something that looks
-			 reasonably like a valid header.  This is not a
-			 perfect strategy, but it should get us back on the
-			 track within a short time (and hopefully without
-			 too much distortion in the audio output).  */
-			do
-			{
-				++try;
-				if(limit >= 0 && try >= limit) break;				
-
-				if((ret=fr->rd->head_shift(fr,&newhead)) <= 0)
-				{
-					debug("need more?");
-					if(NOQUIET) fprintf (stderr, "Note: Hit end of (available) data during resync.\n");
-
-					goto read_frame_bad;
-				}
-				if(VERBOSE3) debug3("resync try %li at %"OFF_P", got newhead 0x%08lx", try, (off_p)fr->rd->tell(fr),  newhead);
-
-				if(!fr->oldhead)
-				{
-					debug("going to init_resync...");
-					goto init_resync;       /* "considered harmful", eh? */
-				}
-				/* we should perhaps collect a list of valid headers that occured in file... there can be more */
-				/* Michael's new resync routine seems to work better with the one frame readahead (and some input buffering?) */
-			} while
-			(
-				!head_check(newhead) /* Simply check for any valid header... we have the readahead to get it straight now(?) */
-				/*   (newhead & HDRCMPMASK) != (fr->oldhead & HDRCMPMASK)
-				&& (newhead & HDRCMPMASK) != (fr->firsthead & HDRCMPMASK)*/
-			);
-			/* too many false positives 
-			}while (!(head_check(newhead) && decode_header(fr, newhead))); */
-			if(NOQUIET && fr->silent_resync == 0) fprintf (stderr, "Note: Skipped %li bytes in input.\n", try);
-
-			if(limit >= 0 && try >= limit)
-			{
-				if(NOQUIET)
-				error1("Giving up resync after %li bytes - your stream is not nice... (maybe increasing resync limit could help).", try);
-
-				fr->err = MPG123_RESYNC_FAIL;
-				return READER_ERROR;
-			}
-			else
-			{
-				debug1("Found valid header 0x%lx... unsetting firsthead to reinit stream.", newhead);
-				fr->firsthead = 0;
-				goto init_resync;
-			}
-		}
-		else
-		{
-			if(NOQUIET) error("not attempting to resync...");
-
-			fr->err = MPG123_OUT_OF_SYNC;
-			return READER_ERROR;
-		}
+		ret = do_readahead(fr, newhead);
+		/* readahead can fail mit NEED_MORE, in which case we must also make the just read header available again for next go */
+		if(ret < 0) fr->rd->back_bytes(fr, 4);
+		JUMP_CONCLUSION(ret);
 	}
-	/* There used to be code here ... and turned out to be totally redundant. */
 
-	/* Past this point we know the header is good and we can start reading in the actual frame. */
+	/* Now we should have our valid header and proceed to reading the frame. */
 
 	/* if filepos is invalid, so is framepos */
 	framepos = fr->rd->tell(fr) - 4;
@@ -775,19 +568,29 @@ init_resync:
 	++fr->num; /* 0 for first frame! */
 	debug4("Frame %"OFF_P" %08lx %i, next filepos=%"OFF_P, 
 	(off_p)fr->num, newhead, fr->framesize, (off_p)fr->rd->tell(fr));
-
-	/* save for repetition */
-	if(fr->p.halfspeed && fr->lay == 3)
+	if(!(fr->state_flags & FRAME_FRANKENSTEIN) && (
+		(fr->track_frames > 0 && fr->num >= fr->track_frames)
+#ifdef GAPLESS
+		|| (fr->gapless_frames > 0 && fr->num >= fr->gapless_frames)
+#endif
+	))
 	{
-		debug("halfspeed - reusing old bsbuf ");
-		memcpy (fr->ssave, fr->bsbuf, fr->ssize);
+		fr->state_flags |= FRAME_FRANKENSTEIN;
+		if(NOQUIET) fprintf(stderr, "\nWarning: Encountered more data after announced end of track (frame %"OFF_P"/%"OFF_P"). Frankenstein!\n", (off_p)fr->num, 
+#ifdef GAPLESS
+		fr->gapless_frames > 0 ? (off_p)fr->gapless_frames : 
+#endif
+		(off_p)fr->track_frames);
 	}
 
+	halfspeed_prepare(fr);
+
 	/* index the position */
+	fr->input_offset = framepos;
 #ifdef FRAME_INDEX
 	/* Keep track of true frame positions in our frame index.
 	   but only do so when we are sure that the frame number is accurate... */
-	if(fr->accurate && FI_NEXT(fr->index, fr->num))
+	if((fr->state_flags & FRAME_ACCURATE) && FI_NEXT(fr->index, fr->num))
 	fi_add(&fr->index, framepos);
 #endif
 
@@ -798,9 +601,11 @@ init_resync:
 	fr->to_decode = fr->to_ignore = TRUE;
 	if(fr->error_protection) fr->crc = getbits(fr, 16); /* skip crc */
 
+	fr->oldhead = newhead;
+
 	return 1;
 read_frame_bad:
-	/* Also if we searched for valid data in vain, we can forget skipped data.
+	/* Also if we searched for valid data in vein, we can forget skipped data.
 	   Otherwise, the feeder would hold every dead old byte in memory until the first valid frame! */
 	if(fr->rd->forget != NULL) fr->rd->forget(fr);
 
@@ -835,38 +640,17 @@ static int guess_freeformat_framesize(mpg123_handle *fr)
 	return ret;
 
 	/* We are already 4 bytes into it */
-/* fix that limit to be absolute for the first header search! */
-	for(i=4;i<65536;i++) {
-		if((ret=fr->rd->head_shift(fr,&head))<=0)
+	/* Fix that limit to be absolute for the first header search! */
+	for(i=4;i<65536;i++)
+	{
+		if((ret=fr->rd->head_shift(fr,&head))<=0) return ret;
+
+		/* No head_check needed, the mask contains all relevant bits. */
+		if((head & HDR_SAMEMASK) == (fr->oldhead & HDR_SAMEMASK))
 		{
-			return ret;
-		}
-		if(head_check(head))
-		{
-			int sampling_frequency,mpeg25,lsf;
-			
-			if(head & (1<<20))
-			{
-				lsf = (head & (1<<19)) ? 0x0 : 0x1;
-				mpeg25 = 0;
-			}
-			else
-			{
-				lsf = 1;
-				mpeg25 = 1;
-			}
-			
-			if(mpeg25)
-				sampling_frequency = 6 + ((head>>10)&0x3);
-			else
-				sampling_frequency = ((head>>10)&0x3) + (lsf*3);
-			
-			if((lsf==fr->lsf) && (mpeg25==fr->mpeg25) && (sampling_frequency == fr->sampling_frequency))
-			{
-				fr->rd->back_bytes(fr,i+1);
-				fr->framesize = i-3;
-				return 1; /* Success! */
-			}
+			fr->rd->back_bytes(fr,i+1);
+			fr->framesize = i-3;
+			return 1; /* Success! */
 		}
 	}
 	fr->rd->back_bytes(fr,i);
@@ -881,18 +665,19 @@ static int guess_freeformat_framesize(mpg123_handle *fr)
  *  1: success
  *  0: no valid header
  * <0: some error
+ * You are required to do a head_check() before calling!
  */
 static int decode_header(mpg123_handle *fr,unsigned long newhead)
 {
+#ifdef DEBUG /* Do not waste cycles checking the header twice all the time. */
 	if(!head_check(newhead))
 	{
-		if(NOQUIET) error("tried to decode obviously invalid header");
-
-		return 0;
+		error1("trying to decode obviously invalid header 0x%08lx", newhead);
 	}
-	if( newhead & (1<<20) )
+#endif
+	if(HDR_VERSION_VAL(newhead) & 0x2)
 	{
-		fr->lsf = (newhead & (1<<19)) ? 0x0 : 0x1;
+		fr->lsf = (HDR_VERSION_VAL(newhead) & 0x1) ? 0 : 1;
 		fr->mpeg25 = 0;
 	}
 	else
@@ -902,43 +687,36 @@ static int decode_header(mpg123_handle *fr,unsigned long newhead)
 	}
 
 	if(   (fr->p.flags & MPG123_NO_RESYNC) || !fr->oldhead
-	   || (((fr->oldhead>>19)&0x3) ^ ((newhead>>19)&0x3))  )
+	   || (HDR_VERSION_VAL(fr->oldhead) != HDR_VERSION_VAL(newhead)) )
 	{
 		/* If "tryresync" is false, assume that certain
 		parameters do not change within the stream!
 		Force an update if lsf or mpeg25 settings
 		have changed. */
-		fr->lay = 4-((newhead>>17)&3);
-		if( ((newhead>>10)&0x3) == 0x3)
-		{
-			if(NOQUIET) error("Stream error");
-
-			return 0; /* exit() here really is too much, isn't it? */
-		}
+		fr->lay = 4 - HDR_LAYER_VAL(newhead);
 		if(fr->mpeg25)
-		fr->sampling_frequency = 6 + ((newhead>>10)&0x3);
+		fr->sampling_frequency = 6 + HDR_SAMPLERATE_VAL(newhead);
 		else
-		fr->sampling_frequency = ((newhead>>10)&0x3) + (fr->lsf*3);
+		fr->sampling_frequency = HDR_SAMPLERATE_VAL(newhead) + (fr->lsf*3);
 	}
 
 	#ifdef DEBUG
-	if((((newhead>>16)&0x1)^0x1) != fr->error_protection) debug("changed crc bit!");
+	/* seen a file where this varies (old lame tag without crc, track with crc) */
+	if((HDR_CRC_VAL(newhead)^0x1) != fr->error_protection) debug("changed crc bit!");
 	#endif
-	fr->error_protection = ((newhead>>16)&0x1)^0x1; /* seen a file where this varies (old lame tag without crc, track with crc) */
-	fr->bitrate_index = ((newhead>>12)&0xf);
-	fr->padding   = ((newhead>>9)&0x1);
-	fr->extension = ((newhead>>8)&0x1);
-	fr->mode      = ((newhead>>6)&0x3);
-	fr->mode_ext  = ((newhead>>4)&0x3);
-	fr->copyright = ((newhead>>3)&0x1);
-	fr->original  = ((newhead>>2)&0x1);
-	fr->emphasis  = newhead & 0x3;
-	fr->freeformat = free_format_header(newhead);
+	fr->error_protection = HDR_CRC_VAL(newhead)^0x1;
+	fr->bitrate_index    = HDR_BITRATE_VAL(newhead);
+	fr->padding          = HDR_PADDING_VAL(newhead);
+	fr->extension        = HDR_PRIVATE_VAL(newhead);
+	fr->mode             = HDR_CHANNEL_VAL(newhead);
+	fr->mode_ext         = HDR_CHANEX_VAL(newhead);
+	fr->copyright        = HDR_COPYRIGHT_VAL(newhead);
+	fr->original         = HDR_ORIGINAL_VAL(newhead);
+	fr->emphasis         = HDR_EMPHASIS_VAL(newhead);
+	fr->freeformat       = !(newhead & HDR_BITRATE);
 
-	fr->stereo    = (fr->mode == MPG_MD_MONO) ? 1 : 2;
+	fr->stereo = (fr->mode == MPG_MD_MONO) ? 1 : 2;
 
-	fr->oldhead = newhead;
-	
 	/* we can't use tabsel_123 for freeformat, so trying to guess framesize... */
 	if(fr->freeformat)
 	{
@@ -957,8 +735,9 @@ static int decode_header(mpg123_handle *fr,unsigned long newhead)
 			{
 				if(ret == MPG123_NEED_MORE)
 				debug("Need more data to guess free format frame size.");
-				else
+				else if(VERBOSE3)
 				error("Encountered free format header, but failed to guess frame size.");
+
 				return ret;
 			}
 		}
@@ -1062,6 +841,13 @@ double compute_bpf(mpg123_handle *fr)
 	return bpf;
 }
 
+int attribute_align_arg mpg123_spf(mpg123_handle *mh)
+{
+	if(mh == NULL) return MPG123_ERR;
+
+	return spf(mh);
+}
+
 double attribute_align_arg mpg123_tpf(mpg123_handle *fr)
 {
 	static int bs[4] = { 0,384,1152,1152 };
@@ -1155,3 +941,234 @@ int get_songlen(mpg123_handle *fr,int no)
 	tpf = mpg123_tpf(fr);
 	return (int) (no*tpf);
 }
+
+/* first attempt of read ahead check to find the real first header; cannot believe what junk is out there! */
+static int do_readahead(mpg123_handle *fr, unsigned long newhead)
+{
+	unsigned long nexthead = 0;
+	int hd = 0;
+	off_t start, oret;
+	int ret;
+
+	if( ! (!fr->firsthead && fr->rdat.flags & (READER_SEEKABLE|READER_BUFFERED)) )
+	return PARSE_GOOD;
+
+	start = fr->rd->tell(fr);
+
+	debug2("doing ahead check with BPF %d at %"OFF_P, fr->framesize+4, (off_p)start);
+	/* step framesize bytes forward and read next possible header*/
+	if((oret=fr->rd->skip_bytes(fr, fr->framesize))<0)
+	{
+		if(oret==READER_ERROR && NOQUIET) error("cannot seek!");
+
+		return oret == MPG123_NEED_MORE ? PARSE_MORE : PARSE_ERR;
+	}
+
+	/* Read header, seek back. */
+	hd = fr->rd->head_read(fr,&nexthead);
+	if( fr->rd->back_bytes(fr, fr->rd->tell(fr)-start) < 0 )
+	{
+		if(NOQUIET) error("Cannot seek back!");
+
+		return PARSE_ERR;
+	}
+	if(hd == MPG123_NEED_MORE) return PARSE_MORE;
+
+	debug1("After fetching next header, at %"OFF_P, (off_p)fr->rd->tell(fr));
+	if(!hd)
+	{
+		if(NOQUIET) warning("Cannot read next header, a one-frame stream? Duh...");
+		return PARSE_END;
+	}
+
+	debug2("does next header 0x%08lx match first 0x%08lx?", nexthead, newhead);
+	if(!head_check(nexthead) || (nexthead & HDR_CMPMASK) != (newhead & HDR_CMPMASK))
+	{
+		debug("No, the header was not valid, start from beginning...");
+		fr->oldhead = 0; /* start over */
+		/* try next byte for valid header */
+		if((ret=fr->rd->back_bytes(fr, 3))<0)
+		{
+			if(NOQUIET) error("Cannot seek 3 bytes back!");
+
+			return PARSE_ERR;
+		}
+		return PARSE_AGAIN;
+	}
+	else return PARSE_GOOD;
+}
+
+static int handle_id3v2(mpg123_handle *fr, unsigned long newhead)
+{
+	int ret;
+	fr->oldhead = 0; /* Think about that. Used to be present only for skipping of junk, not resync-style wetwork. */
+	ret = parse_new_id3(fr, newhead);
+	if     (ret < 0) return ret;
+#ifndef NO_ID3V2
+	else if(ret > 0){ debug("got ID3v2"); fr->metaflags  |= MPG123_NEW_ID3|MPG123_ID3; }
+	else debug("no useful ID3v2");
+#endif
+	return PARSE_AGAIN;
+}
+
+/* watch out for junk/tags on beginning of stream by invalid header */
+static int skip_junk(mpg123_handle *fr, unsigned long *newheadp, long *headcount)
+{
+	int ret;
+	long limit = 65536;
+	unsigned long newhead = *newheadp;
+	/* check for id3v2; first three bytes (of 4) are "ID3" */
+	if((newhead & (unsigned long) 0xffffff00) == (unsigned long) 0x49443300)
+	{
+		return handle_id3v2(fr, newhead);
+	}
+	else if(VERBOSE2 && fr->silent_resync == 0) fprintf(stderr,"Note: Junk at the beginning (0x%08lx)\n",newhead);
+
+	/* I even saw RIFF headers at the beginning of MPEG streams ;( */
+	if(newhead == ('R'<<24)+('I'<<16)+('F'<<8)+'F')
+	{
+		if(VERBOSE2 && fr->silent_resync == 0) fprintf(stderr, "Note: Looks like a RIFF header.\n");
+
+		if((ret=fr->rd->head_read(fr,&newhead))<=0) return ret;
+
+		while(newhead != ('d'<<24)+('a'<<16)+('t'<<8)+'a')
+		{
+			if((ret=fr->rd->head_shift(fr,&newhead))<=0) return ret;
+		}
+		if((ret=fr->rd->head_read(fr,&newhead))<=0) return ret;
+
+		if(VERBOSE2 && fr->silent_resync == 0) fprintf(stderr,"Note: Skipped RIFF header!\n");
+
+		fr->oldhead = 0;
+		*newheadp = newhead;
+		return PARSE_AGAIN;
+	}
+
+	/*
+		Unhandled junk... just continue search for a header, stepping in single bytes through next 64K.
+		This is rather identical to the resync loop.
+	*/
+	debug("searching for header...");
+	*newheadp = 0; /* Invalidate the external value. */
+	ret = 0; /* We will check the value after the loop. */
+
+	/* We prepare for at least the 64K bytes as usual, unless
+	   user explicitly wanted more (even infinity). Never less. */
+	if(fr->p.resync_limit < 0 || fr->p.resync_limit > limit)
+	limit = fr->p.resync_limit;
+
+	do
+	{
+		++(*headcount);
+		if(limit >= 0 && *headcount >= limit) break;				
+
+		if((ret=fr->rd->head_shift(fr,&newhead))<=0) return ret;
+
+		if(head_check(newhead) && (ret=decode_header(fr, newhead))) break;
+	} while(1);
+	if(ret<0) return ret;
+
+	if(limit >= 0 && *headcount >= limit)
+	{
+		if(NOQUIET) error1("Giving up searching valid MPEG header after %li bytes of junk.", *headcount);
+		return PARSE_END;
+	}
+	else debug1("hopefully found one at %"OFF_P, (off_p)fr->rd->tell(fr));
+
+	/* If the new header ist good, it is already decoded. */
+	*newheadp = newhead;
+	return PARSE_GOOD;
+}
+
+/* The newhead is bad, so let's check if it is something special, otherwise just resync. */
+static int wetwork(mpg123_handle *fr, unsigned long *newheadp)
+{
+	int ret = PARSE_ERR;
+	unsigned long newhead = *newheadp;
+	*newheadp = 0;
+
+	/* Classic ID3 tags. Read, then start parsing again. */
+	if((newhead & 0xffffff00) == ('T'<<24)+('A'<<16)+('G'<<8))
+	{
+		fr->id3buf[0] = (unsigned char) ((newhead >> 24) & 0xff);
+		fr->id3buf[1] = (unsigned char) ((newhead >> 16) & 0xff);
+		fr->id3buf[2] = (unsigned char) ((newhead >> 8)  & 0xff);
+		fr->id3buf[3] = (unsigned char) ( newhead        & 0xff);
+
+		if((ret=fr->rd->fullread(fr,fr->id3buf+4,124)) < 0) return ret;
+
+		fr->metaflags  |= MPG123_NEW_ID3|MPG123_ID3;
+		fr->rdat.flags |= READER_ID3TAG; /* that marks id3v1 */
+		if(VERBOSE3) fprintf(stderr,"Note: Skipped ID3v1 tag.\n");
+
+		return PARSE_AGAIN;
+	}
+	/* This is similar to initial junk skipping code... */
+	/* Check for id3v2; first three bytes (of 4) are "ID3" */
+	if((newhead & (unsigned long) 0xffffff00) == (unsigned long) 0x49443300)
+	{
+		return handle_id3v2(fr, newhead);
+	}
+	else if(NOQUIET && fr->silent_resync == 0)
+	{
+		fprintf(stderr,"Note: Illegal Audio-MPEG-Header 0x%08lx at offset %"OFF_P".\n",
+			newhead, (off_p)fr->rd->tell(fr)-4);
+	}
+
+	/* Now we got something bad at hand, try to recover. */
+
+	if(NOQUIET && (newhead & 0xffffff00) == ('b'<<24)+('m'<<16)+('p'<<8)) fprintf(stderr,"Note: Could be a BMP album art.\n");
+
+	if( !(fr->p.flags & MPG123_NO_RESYNC) )
+	{
+		long try = 0;
+		long limit = fr->p.resync_limit;
+
+		/* If a resync is needed the bitreservoir of previous frames is no longer valid */
+		fr->bitreservoir = 0;
+
+		if(NOQUIET && fr->silent_resync == 0) fprintf(stderr, "Note: Trying to resync...\n");
+
+		do /* ... shift the header with additional single bytes until be found something that could be a header. */
+		{
+			++try;
+			if(limit >= 0 && try >= limit) break;				
+
+			if((ret=fr->rd->head_shift(fr,&newhead)) <= 0)
+			{
+				*newheadp = newhead;
+				if(NOQUIET) fprintf (stderr, "Note: Hit end of (available) data during resync.\n");
+
+				return ret;
+			}
+			if(VERBOSE3) debug3("resync try %li at %"OFF_P", got newhead 0x%08lx", try, (off_p)fr->rd->tell(fr),  newhead);
+		} while(!head_check(newhead));
+
+		*newheadp = newhead;
+		if(NOQUIET && fr->silent_resync == 0) fprintf (stderr, "Note: Skipped %li bytes in input.\n", try);
+
+		/* Now we either got something that could be a header, or we gave up. */
+		if(limit >= 0 && try >= limit)
+		{
+			if(NOQUIET)
+			error1("Giving up resync after %li bytes - your stream is not nice... (maybe increasing resync limit could help).", try);
+
+			fr->err = MPG123_RESYNC_FAIL;
+			return PARSE_ERR;
+		}
+		else
+		{
+			debug1("Found possibly valid header 0x%lx... unsetting firsthead to reinit stream.", newhead);
+			fr->firsthead = 0;
+			return PARSE_RESYNC;
+		}
+	}
+	else
+	{
+		if(NOQUIET) error("not attempting to resync...");
+
+		fr->err = MPG123_OUT_OF_SYNC;
+		return PARSE_ERR;
+	}
+	/* Control never goes here... we return before that. */
+}
